package engine

import "C"
import (
	"bytes"
	"compress/flate"
	"compress/gzip"
	"context"
	"crypto/tls"
	"errors"
	"fmt"
	"io"
	"log"
	// "mercury/engine/httpntlm"
	"mercury/engine/spnego"
	"net"
	"net/http"
	"net/http/httptrace"
	"net/http/httputil"
	"net/url"
	"os"
	"path"
	"strconv"
	"strings"
	"sync"
	"time"
	"unsafe"

	// "github.com/MarshallWace/go-spnego"
	"github.com/andybalholm/brotli"
	"github.com/gregjones/httpcache"
	"github.com/icholy/digest"
)

type response_check_results struct {
	failed_web_reg_save_param bool
	failed_web_reg_find       bool
	failed_http_status        bool

	failed_web_reg_save_param_items map[string]bool
	failed_web_reg_find_items       map[string]bool
}

type extra_res struct {
	Url     string
	Referer string
}

type web_request struct {
	ThreadId               string
	RequestName            string
	RequestId              string
	URL                    string
	Method                 string         `yaml:",omitempty"`
	TargetFrame            string         `yaml:",omitempty"`
	EncType                string         `yaml:",omitempty"`
	RecContentType         string         `yaml:",omitempty"`
	Referer                string         `yaml:",omitempty"`
	Body                   string         `yaml:",omitempty"`
	Resource               int            `yaml:",omitempty"`
	Snapshot               string         `yaml:",omitempty"`
	Mode                   string         `yaml:",omitempty"`
	UserAgent              string         `yaml:",omitempty"`
	ContentEncoding        string         `yaml:",omitempty"`
	EncodeAtSign           bool           `yaml:",omitempty"`
	Headers                http.Header    `yaml:",omitempty"`
	Cookies                []*http.Cookie `yaml:",omitempty"`
	response_results       []response_check_results
	response_results_mutex sync.Mutex
	is_concurrent          bool
	is_async               bool
	redirectURL            string
	extraResources         []extra_res
	async_attribs          *async_attributes
	cancel                 context.CancelFunc

	RawRequest  strings.Builder `yaml:",omitempty"`
	RawResponse strings.Builder `yaml:",omitempty"`
}

var active_requests map[string]*web_request = make(map[string]*web_request)
var active_requests_mutex sync.Mutex

var ExceededMaxRedirectionDepthError = errors.New("Exceeded MaxRedirectionDepth")

func get_active_request(requestId string) *web_request {
	active_requests_mutex.Lock()
	defer active_requests_mutex.Unlock()
	req, ok := active_requests[requestId]
	if ok {
		return req
	} else {
		return nil
	}
}

//export web_custom_request_exec
func web_custom_request_exec(requestName *C.char, args **C.char, args_len C.int) C.int {
	arg_list := (*[1 << 30]*C.char)(unsafe.Pointer(args))[:args_len:args_len]

	var sarg_list []string
	for _, v := range arg_list {
		sarg_list = append(sarg_list, C.GoString(v))
	}

	return C.int(go_web_custom_request_exec(C.GoString(requestName), sarg_list, int(args_len)))
}

// type TransportMiddleware struct {
// 	http.Transport
// 	RoundTripper *http.RoundTripper
// }

// func (t *TransportMiddleware) RoundTrip(req *http.Request) (*http.Response, error) {
// 	if t.RoundTripper != nil {
// 		return t.RoundTripper.RoundTrip(req)
// 	}
// 	return nil, errors.New("RoundTripper not set")
// }

func go_web_custom_request_exec(requestName string, args []string, args_len int) int {
	wt_end := track_wasted_time()
	// vlog("web_custom_request!")

	var req web_request
	req.ThreadId = threadID()
	req.RequestName = go_lr_eval_string(requestName)
	req.RequestId = fmt.Sprintf("%s_%d_%s", threadID(), time.Now().UnixNano(), req.RequestName)
	req.Method = "GET"
	req.Mode = "HTTP"

	extraResMode := false
	var extraRes extra_res

	for _, value := range args {
		arg := go_lr_eval_string(value)

		// if vuser().runtimeSettings.Log.AdvancedTrace {
		// 	vlog("web_custom_request! arg=", arg)
		// }

		if arg == EXTRARES {
			extraResMode = true
			extraRes = extra_res{Referer: req.URL}
			continue
		} else if arg == ENDITEM && extraResMode {
			req.extraResources = append(req.extraResources, extraRes)
			extraRes = extra_res{Referer: req.URL}
			continue
		}

		namevalue := strings.SplitN(arg, "=", 2)
		name := namevalue[0]
		value := namevalue[1]

		// if RuntimeSettings.Log.AdvancedTrace {
		// 	vlog("web_custom_request! name=", name, "value=", value)
		// }

		if !extraResMode {
			if strings.ToUpper(name) == "URL" {
				req.URL = SafeURL(value)
			} else if name == "Method" {
				req.Method = value
			} else if name == "Body" {
				req.Body = value
			} else if name == "BodyBinary" {
				req.Body = value
			} else if name == "BodyUnicode" {
				req.Body = value
			} else if name == "Resource" {
				resource, err := strconv.Atoi(value)
				if err == nil {
					req.Resource = resource
				}
			} else if name == "RecContentType" {
				req.RecContentType = value
			} else if name == "EncType" {
				req.EncType = value
			} else if name == "Mode" {
				req.Mode = value
			} else if name == "Referer" {
				req.Referer = value
			} else if name == "EncodeAtSign" {
				req.EncodeAtSign = (strings.ToUpper(value) == "YES")
			}
		} else {
			if strings.ToUpper(name) == "URL" {
				extraRes.Url = value
			} else if name == "Referer" {
				extraRes.Referer = value
			}
		}
	}

	wt_end()

	if web_concurrent_request(&req) {
		return LR_PASS
	}

	return web_custom_request_exec_impl(req)
}

// Create a static http client (i.e. certificate details go here)
var httpClientStore map[string]*http.Client
var httpClientStore_mutex sync.Mutex

func get_http_client() *http.Client {
	httpClientStore_mutex.Lock()
	defer httpClientStore_mutex.Unlock()

	if httpClientStore == nil {
		httpClientStore = make(map[string]*http.Client)
	}

	id := threadID()

	if httpClientStore[id] == nil {
		vlog("Initialize http client")

		RuntimeSettings := vuser().runtimeSettings

		proxy := http.ProxyFromEnvironment
		if RuntimeSettings.Proxy.Mode == 0 {
			proxy = nil
		} else if RuntimeSettings.Proxy.Mode == 1 {
			proxy = http.ProxyFromEnvironment
		} else if RuntimeSettings.Proxy.Mode == 2 {
			creds := ""
			if RuntimeSettings.Proxy.Username != "" && RuntimeSettings.Proxy.Password != "" {
				creds = fmt.Sprintf("%s:%s@", RuntimeSettings.Proxy.Username, RuntimeSettings.Proxy.Password)
			}

			if RuntimeSettings.Proxy.HttpsProxy != "" {
				https_proxy_url := fmt.Sprintf("https://%s%s:%d", creds, RuntimeSettings.Proxy.HttpsProxy, RuntimeSettings.Proxy.HttpsProxyPort)
				vlog("HTTPS Proxy URL Set:", https_proxy_url)
				proxyURL, err := url.Parse(https_proxy_url)
				if err == nil {
					proxy = http.ProxyURL(proxyURL)
				}
			} else if RuntimeSettings.Proxy.HttpProxy != "" {
				http_proxy_url := fmt.Sprintf("http://%s%s:%d", creds, RuntimeSettings.Proxy.HttpProxy, RuntimeSettings.Proxy.HttpProxyPort)
				vlog("HTTP Proxy URL Set:", http_proxy_url)
				proxyURL, err := url.Parse(http_proxy_url)
				if err == nil {
					proxy = http.ProxyURL(proxyURL)
				}
			}
		}

		tlsConfig := &tls.Config{
			//RootCAs:      caCertPool,
			InsecureSkipVerify: true,
		}

		dialer := &net.Dialer{
			Timeout:   time.Duration(RuntimeSettings.Preferences.HTTPRequestConnectTimeout) * time.Second,
			KeepAlive: time.Duration(RuntimeSettings.Preferences.HTTPKeepAliveTimeout) * time.Second,
		}

		var transport http.RoundTripper

		mainTransport := http.Transport{
			TLSClientConfig:       tlsConfig,
			TLSHandshakeTimeout:   time.Duration(RuntimeSettings.Preferences.HTTPRequestConnectTimeout) * time.Second,
			IdleConnTimeout:       time.Duration(RuntimeSettings.Preferences.HTTPKeepAliveTimeout) * time.Second,
			ResponseHeaderTimeout: time.Duration(RuntimeSettings.Preferences.HTTPRequestReceiveTimeout) * time.Second,
			// Dial: (&net.Dialer{
			// 	Timeout:   time.Duration(RuntimeSettings.Preferences.HTTPRequestConnectTimeout) * time.Second,
			// 	KeepAlive: time.Duration(RuntimeSettings.Preferences.HTTPKeepAliveTimeout) * time.Second,
			// }).Dial,
			DialContext: func(ctx context.Context, network, address string) (net.Conn, error) {
				c, err := dialer.Dial(network, address)
				if err != nil {
					return nil, err
				}

				var out io.Writer = os.Stdout
				vu, ok := vuserById(id)
				if ok && vu.log != nil {
					out = vu.log.Writer()
				}

				requestId, ok := ctx.Value("requestId").(string)
				if !ok {
					return c, nil
				}

				req_spec := get_active_request(requestId)

				go_lr_log_vuser(id, fmt.Sprint("DIAL!!", requestId))

				// vlogf("\n%s\n", strings.Repeat("-", 80))
				return WrapConnection(c, out, req_spec), nil
			},
			DialTLSContext: func(ctx context.Context, network, address string) (net.Conn, error) {
				plainConn, err := dialer.Dial(network, address)
				if err != nil {
					return nil, err
				}

				//Initiate TLS and check remote host name against certificate.
				cfg := new(tls.Config)
				cfg.InsecureSkipVerify = true

				// add https:// to satisfy url.Parse(), we won't use it
				u, err := url.Parse(fmt.Sprintf("https://%s", address))
				if err != nil {
					return nil, err
				}

				serverName := u.Host[:strings.LastIndex(u.Host, ":")]
				cfg.ServerName = serverName

				tlsConn := tls.Client(plainConn, cfg)

				// errc := make(chan error, 2)
				// timer := time.AfterFunc(time.Duration(RuntimeSettings.Preferences.HTTPRequestConnectTimeout)*time.Second, func() {
				// 	errc <- errors.New("TLS handshake timeout")
				// })
				// go func() {
				// 	err := tlsConn.Handshake()
				// 	timer.Stop()
				// 	errc <- err
				// }()
				// if err := <-errc; err != nil {
				// 	plainConn.Close()
				// 	return nil, err
				// }
				// if !cfg.InsecureSkipVerify {
				// 	if err := tlsConn.VerifyHostname(cfg.ServerName); err != nil {
				// 		plainConn.Close()
				// 		return nil, err
				// 	}
				// }

				var out io.Writer = os.Stdout
				vu, ok := vuserById(id)
				if ok && vu.log != nil {
					out = vu.log.Writer()
				}

				requestId, ok := ctx.Value("requestId").(string)
				if !ok {
					return tlsConn, nil
				}

				req_spec := get_active_request(requestId)
				// vlogf("\n%s\n", strings.Repeat("-", 80))
				return WrapConnection(tlsConn, out, req_spec), nil
			},
			Proxy: proxy,
		}

		// if runtime setting has cache enabled
		isCacheEnabled := RuntimeSettings.Browser.EnableCache
		if isCacheEnabled {
			cache := *init_cache()

			// Create a new HTTP transport with caching enabled
			cacheTransport := httpcache.NewTransport(cache)
			cacheTransport.Transport = &mainTransport

			transport = cacheTransport
		} else {
			transport = &mainTransport
		}

		cookieJar := get_cookie_jar()

		transport = &digest.Transport{
			Transport: transport,
			Jar:       cookieJar,
			Digest: func(req *http.Request, chal *digest.Challenge, opt digest.Options) (*digest.Credentials, error) {
				username, password, ok := add_digestauth_to_request(req)
				if ok {
					opt.Username = username
					opt.Password = password
				}
				return digest.Digest(chal, opt)
			},
		}

		// transport = &httpntlm.NtlmTransport{
		// 	RoundTripper: transport,
		// 	Jar:          cookieJar,
		// 	GetCreds: func(req *http.Request) (string, string, string) {
		// 		domain, username, password, _ := add_ntlmauth_to_request(req)
		// 		return domain, username, password
		// 	},
		// 	ModifyRequest: func(req *http.Request) {
		// 		requestId, ok := req.Context().Value("requestId").(string)
		// 		if ok {
		// 			req_spec := get_active_request(requestId)

		// 			if req_spec.Referer != "" {
		// 				req.Header.Set("Referer", req_spec.Referer)
		// 			}
		// 		}
				
		// 		req.Header.Set("Accept-Encoding", RuntimeSettings.Preferences.AcceptServerSideCompression)
		// 		req.Header.Set("User-Agent", RuntimeSettings.Browser.UserAgent)
		// 		req.Header.Set("Accept-Language", "en-US")
		// 		req.Header.Set("Accept", "*/*")
		// 		add_headers_to_request(req)
		// 	},
		// 	OnResponse: func(req *http.Request, res *http.Response) {
		// 		requestId, ok := req.Context().Value("requestId").(string)
		// 		vlog("OnNTLMResponse",requestId,ok)
		// 		if ok {
		// 			req_spec := get_active_request(requestId)
		// 			defer req_spec.RawResponse.Reset()
		// 			vlog("NTLMResponse:",req_spec.RawResponse.String())
		// 		}
		// 	},
		// }

		transport = &spnego.Transport{
			RoundTripper: transport,
			Jar:          cookieJar,
			GetCreds: func(req *http.Request) (string, string, string) {
				domain, username, password, _ := add_ntlmauth_to_request(req)
				return domain, username, password
			},
			ModifyRequest: func(req *http.Request) {
				requestId, ok := req.Context().Value("requestId").(string)
				if ok {
					req_spec := get_active_request(requestId)

					if req_spec.Referer != "" {
						req.Header.Set("Referer", req_spec.Referer)
					}
				}
				
				req.Header.Set("Accept-Encoding", RuntimeSettings.Preferences.AcceptServerSideCompression)
				req.Header.Set("User-Agent", RuntimeSettings.Browser.UserAgent)
				req.Header.Set("Accept-Language", "en-US")
				req.Header.Set("Accept", "*/*")
				add_headers_to_request(req)
			},
			OnResponse: func(req *http.Request, res *http.Response) {
				requestId, ok := req.Context().Value("requestId").(string)
				vlog("OnSpnegoResponse",requestId,ok)
				if ok {
					req_spec := get_active_request(requestId)
					defer req_spec.RawResponse.Reset()
					vlog("Spnego Response:",req_spec.RawResponse.String())
				}
			},
		}

		httpClientStore[id] = &http.Client{
			Timeout: time.Duration(RuntimeSettings.Preferences.HTTPRequestConnectTimeout+
				RuntimeSettings.Preferences.HTTPRequestReceiveTimeout) * time.Second,
			Transport:     transport,
			Jar:           cookieJar,
			CheckRedirect: redirect_handler,
		}
	}

	return httpClientStore[id]
}

func redirect_handler(req *http.Request, via []*http.Request) error {
	//return http.ErrUseLastResponse

	RuntimeSettings := vuser().runtimeSettings
	redirect_count := len(via)

	ctx := req.Context()
	requestId := fmt.Sprintf("%s", ctx.Value("requestId"))

	active_requests_mutex.Lock()
	req_spec := active_requests[requestId]
	active_requests_mutex.Unlock()

	web_custom_request_response_handler(req_spec, req.Response)

	if redirect_count > RuntimeSettings.Preferences.MaxRedirectionDepth {
		vlogf("stopped after redirect_count(%d) exceeded MaxRedirectionDepth(%d)", redirect_count, RuntimeSettings.Preferences.MaxRedirectionDepth)
		return ExceededMaxRedirectionDepthError
	}

	safeUrl, safeUrlNeeded := SafeURLQuery(req.URL.String())
	if safeUrlNeeded {
		u, err := url.Parse(safeUrl)
		if err == nil {
			req.URL = u
		}
	}
	req_spec.redirectURL = req.URL.String()

	if RuntimeSettings.Log.AdvancedTrace {
		vlog("Redirect! ", redirect_count, "requestId", requestId, "\nreq=>", req.Method, req.URL)
	}

	add_headers_to_request(req)
	if req.Header.Get("Authorization") == "" {
		add_basicauth_to_request(req)
	}

	return nil
}

func web_custom_request_exec_impl(req_spec web_request) int {
	wt_end := track_wasted_time()
	if !req_spec.is_concurrent {
		defer wt_end()
	}

	if req_spec.RequestId == "" {
		log.Fatal("Request with empty RequestId")
	}

	if req_spec.ThreadId == "" {
		log.Fatal("Request with empty ThreadId")
	}

	RuntimeSettings := vuser().runtimeSettings
	last := clear_lastreq()

	if evaluate_async_attribs(&req_spec) {
		return LR_PASS
	}

	// Ensure this request is in active_requests and removed when this function ends
	active_requests_mutex.Lock()
	active_requests[req_spec.RequestId] = &req_spec
	active_requests_mutex.Unlock()

	defer func() {
		active_requests_mutex.Lock()
		delete(active_requests, req_spec.RequestId)
		active_requests_mutex.Unlock()
	}()

	events().Emit("newrequest_" + req_spec.RequestId)

	// Download Filters
	if web_download_filters_test(&req_spec) == false {
		vlog("[warn] Request cancelled due to download filter:", req_spec.URL)
		return LR_FAIL
	}

	// Automatic Step transactions
	if RuntimeSettings.AutomaticTransactions.Steps && req_spec.Resource == 1 && !req_spec.is_concurrent {
		go_lr_start_transaction(req_spec.RequestName)
		defer go_lr_end_transaction(req_spec.RequestName, LR_AUTO)
	}

	if req_spec.Resource == 1 && RuntimeSettings.Browser.DownloadResources == false {
		return LR_PASS
	}

	// Clear registrations
	defer clear_web_reg_finds()
	defer clear_web_reg_saves()

	// Helper func for exitting this request gracefully on error
	graceful_exit := func() {
		if RuntimeSettings.ErrorHandling.ContinueOnError {
			lr_set_transaction_statuses(LR_FAIL)
			vlog("[warn] Continuing on error..")
		} else {
			lr_exit(LR_EXIT_MAIN_ITERATION_AND_CONTINUE, LR_FAIL)
		}
	}

	vlog("New Request", req_spec.RequestId)

	// Build http request context with values, tracing and a timeout
	ctx := context.Background()
	ctx = context.WithValue(ctx, "threadId", req_spec.ThreadId)
	ctx = context.WithValue(ctx, "requestName", req_spec.RequestName)
	ctx = context.WithValue(ctx, "requestId", req_spec.RequestId)

	clientTrace := httptrace.ClientTrace{
		GotConn: func(info httptrace.GotConnInfo) {
			if RuntimeSettings.Log.AdvancedTrace {
				vlog_vuser(req_spec.ThreadId, "conn was reused:", info.Reused)
				vlogf_vuser(req_spec.ThreadId, "Got Connection Info: %+v", info)
			}
		},
		DNSDone: func(dnsInfo httptrace.DNSDoneInfo) {
			if RuntimeSettings.Log.AdvancedTrace {
				vlogf_vuser(req_spec.ThreadId, "DNS Info: %+v", dnsInfo)
			}
		},
	}
	ctx = httptrace.WithClientTrace(ctx, &clientTrace)

	ctx, req_spec.cancel = context.WithTimeout(ctx, time.Duration(RuntimeSettings.Preferences.StepDownloadTimeout)*time.Second)
	defer req_spec.cancel()

	// Prepare request body payload
	strBody := req_spec.Body
	if req_spec.EncodeAtSign {
		strBody = strings.ReplaceAll(strBody, "@", "&#64;")
	}
	rawBody := []byte(strBody)
	bodyReader := bytes.NewReader(rawBody)

	// Create the request
	req, err := http.NewRequestWithContext(
		ctx,
		strings.TrimSpace(req_spec.Method),
		strings.TrimSpace(req_spec.URL),
		bodyReader,
	)
	if err != nil {
		vlogf_error("client: could not create request: %s", err)
		graceful_exit()
		return LR_FAIL
	}

	// Adding request headers
	if req_spec.Referer != "" {
		req.Header.Set("Referer", req_spec.Referer)
	}
	req.Header.Set("Accept-Encoding", RuntimeSettings.Preferences.AcceptServerSideCompression)
	req.Header.Set("User-Agent", RuntimeSettings.Browser.UserAgent)
	req.Header.Set("Accept-Language", "en-US")
	req.Header.Set("Accept", "*/*")
	add_headers_to_request(req)
	if req.Header.Get("Authorization") == "" {
		add_basicauth_to_request(req)
	}

	// Determine Content Type
	_, hasContentType := req.Header["Content-Type"]
	if req_spec.EncType != "" {
		req.Header.Set("Content-Type", req_spec.EncType)
	} else if !hasContentType && req_spec.Method == "POST" {
		req.Header.Set("Content-Type", "application/x-www-form-urlencoded")
	}

	// Advanced tracing of request details
	// req_spec.Headers = req.Header
	// req_spec.Cookies = get_cookie_jar().Cookies(req.URL)
	web_trace_req_dump(&req_spec, req)

	// Get the HTTP Client
	client := get_http_client()

	if !req_spec.is_concurrent {
		wt_end()
	}

	// Send the request
	res, err := client.Do(req)
	skipResponseHandler := false
	if err != nil && strings.Contains(err.Error(), ExceededMaxRedirectionDepthError.Error()) {
		vlogf("client exceeded MaxRedirectionDepth. Proceeding.")
		skipResponseHandler = true
	} else if err != nil {
		vlogf_error("client: error making http request: %s", err)
		graceful_exit()
		return LR_FAIL
	}
	defer res.Body.Close()

	// Handle the response
	vlog(req_spec.RequestId, "- got response!")
	if !skipResponseHandler && web_custom_request_response_handler(&req_spec, res) == LR_FAIL {
		vlog_error("Failed handling response!", req_spec.RequestId)
		if req_spec.Resource != 1 {
			graceful_exit()
		}
		return LR_FAIL
	}

	vlog("Finished Request!", req_spec.RequestId)
	last.ReturnCode = res.StatusCode

	// Determine overall request status based on checks performed during response handler
	// reg_find_status := LR_FAIL
	// reg_save_status := LR_FAIL
	http_code_status := LR_PASS

	vlog(req_spec.RequestId, "- Response Results", len(req_spec.response_results))
	param_matches := map[string]bool{}
	find_matches := map[string]bool{}
	for _, result := range req_spec.response_results {
		// if !result.failed_web_reg_find {
		// 	reg_find_status = LR_PASS
		// }
		// if !result.failed_web_reg_save_param {
		// 	reg_save_status = LR_PASS
		// }
		if result.failed_http_status {
			http_code_status = LR_FAIL
		}
		param_matches = merge_param_matches(param_matches, result.failed_web_reg_save_param_items)
		find_matches = merge_param_matches(find_matches, result.failed_web_reg_find_items)
	}
	vlog("find_matches",find_matches)

	param_misses := has_param_misses(param_matches)
	find_misses := has_param_misses(find_matches)

	status := LR_FAIL
	if !param_misses && !find_misses && http_code_status == LR_PASS {
		status = LR_PASS
	}

	// Handle non-resource related failures
	if status == LR_FAIL && req_spec.Resource != 1 {
		if find_misses {

			for label, match := range find_matches {
				if !match {
					vlogf_error("web_reg_find: '%s' did not evaluate successfully", label)
				}
			}

			vlog_error(req_spec.RequestId, " - Failed due to 1 or more web_reg_finds")
		}
		if param_misses {

			for paramName, match := range param_matches {
				if !match {
					vlogf_error("web_reg_save: '%s' was not found in response!", paramName)
				}
			}

			vlog_error(req_spec.RequestId, " - Failed due to 1 or more web_reg_save_params")
		}

		graceful_exit()
	}

	// Handle Extra Resources (EXTRARES)
	if req_spec.Resource != 1 && !req_spec.is_concurrent && len(req_spec.extraResources) > 0 {
		res.Body.Close()
		web_custom_request_extrares_handler(&req_spec)
	}

	return status
}

func web_custom_request_response_handler(req *web_request, res *http.Response) int {
	if req == nil || res == nil {
		return LR_FAIL
	}

	defer req.RawResponse.Reset()
	defer req.RawRequest.Reset()

	events().Emit("responseheaders_"+req.RequestId, res)

	// vlog(req.RequestId, "handling response!")
	RuntimeSettings := vuser().runtimeSettings

	var result response_check_results
	result.failed_web_reg_save_param = false
	result.failed_web_reg_find = false
	result.failed_http_status = false
	hasError := false

	if res.StatusCode < 400 {
		vlogf("Response status code: %d (%s)", res.StatusCode, req.RequestId)
	} else {
		result.failed_http_status = true
		vlogf_error("Error response status code: %d (%s)", res.StatusCode, req.RequestId)
		hasError = true
	}

	// Read the response
	var resBody []byte

	// Decompress the response
	decode := RuntimeSettings.Preferences.DecodeContent && (req.Resource != 1 || RuntimeSettings.Preferences.DecodeContentForResources)
	encoding := res.Header.Get("Content-Encoding")
	if decode && encoding == "br" {
		enflated, err := io.ReadAll(brotli.NewReader(res.Body))
		if err != nil {
			vlogf_error("br decompress error: %s", err)
			return LR_FAIL
		}
		resBody = enflated
	} else if decode && encoding == "deflate" {
		enflated, err := io.ReadAll(flate.NewReader(res.Body))
		if err != nil {
			vlogf_error("deflate decompress error: %s", err)
			return LR_FAIL
		}
		resBody = enflated
	} else if decode && (encoding == "gzip" || encoding == "x-gzip") {
		gr, err := gzip.NewReader(res.Body)
		if err != nil {
			vlogf_error("gzip decompress error: %s", err)
			return LR_FAIL
		}
		body, err := io.ReadAll(gr)
		if err != nil {
			vlogf_error("gzip could not read: %s", err)
			//os.Exit(1) // TODO - lr_exit
			return LR_FAIL
		}
		resBody = body
	} else {
		body, err := io.ReadAll(res.Body)
		if err != nil {
			vlogf_error("client: could not read response body: %s", err)
			//os.Exit(1) // TODO - lr_exit
			return LR_FAIL
		}
		resBody = body
	}

	events().Emit("responsebody_"+req.RequestId, map[string]interface{}{
		"resBody": string(resBody),
		"res":     res,
	})

	if req.Resource != 1 || RuntimeSettings.Preferences.DecodeContentForResources {
		wt_end := track_wasted_time()

		rawResponse := req.RawResponse.String()
		endOfHeaders := strings.Index(rawResponse, "\r\n\r\n")

		resHeaders := ""
		if endOfHeaders > 0 {
			// Use headers from raw response
			resHeaders = rawResponse[:endOfHeaders]
		} else {
			// Fallback and use the raw header string function
			resHeaders = raw_header_string(res)
		}

		//vlog("ResHeaders:", resHeaders)

		web_trace_res_dump(req, res, string(resBody), resHeaders)

		body_find_matches := evaluate_web_reg_finds(string(resBody))
		hdr_find_matches := evaluate_web_reg_finds(resHeaders)
		result.failed_web_reg_find_items = merge_param_matches(body_find_matches, hdr_find_matches)

		if has_param_misses(result.failed_web_reg_find_items) {
			result.failed_web_reg_find = true
			hasError = true
			// vlog_error("failed_web_reg_find", req.RequestId)
		}

		body_param_misses := evaluate_web_reg_saves(string(resBody), false, req.Resource == 1)
		vlog("body_param_misses",body_param_misses)
		hdr_param_misses := evaluate_web_reg_saves(resHeaders, true, req.Resource == 1)
		vlog("hdr_param_misses",hdr_param_misses)
		result.failed_web_reg_save_param_items = merge_param_matches(body_param_misses, hdr_param_misses)
		vlog("result.failed_web_reg_save_param_items",result.failed_web_reg_save_param_items)

		if has_param_misses(result.failed_web_reg_save_param_items) {
			result.failed_web_reg_save_param = true
			hasError = true
			// vlog_error("failed_web_reg_save_param", req.RequestId)
		}

		req.response_results_mutex.Lock()
		req.response_results = append(req.response_results, result)
		vlog("finished handling response", len(req.response_results), req.RequestId)
		req.response_results_mutex.Unlock()

		if hasError && RuntimeSettings.ErrorHandling.SnapshotOnError {
			web_snapshot_on_error(req, res, string(resBody))
		}

		wt_end()

		if res.StatusCode < 300 && !req.is_concurrent && req.Mode == "HTML" {
			contentType := res.Header.Get("Content-Type")
			if strings.HasPrefix(contentType, "text/html") || strings.HasPrefix(contentType, "application/html") {
				web_html_handler(req, string(resBody))
			}
		}
	}

	return LR_PASS
}

func web_custom_request_extrares_handler(req *web_request) int {
	wt_end := track_wasted_time()
	defer wt_end()

	base, err := url.Parse(req.URL)
	if err != nil {
		vlog_error("Error parsing base url during extra res handler", err)
		return LR_FAIL
	}

	thread_id := threadID()

	wt_end()
	go_web_concurrent_start(req.RequestName+"_extrares", 0)

	for _, extraRes := range req.extraResources {
		u, _ := url.Parse(extraRes.Url)

		if !u.IsAbs() {
			u = base.ResolveReference(u)
		}

		resourceUrl := u.String()
		name := "extrares_" + path.Base(u.Path)
		requestId := fmt.Sprintf("%s_%d_%s_%s", thread_id, time.Now().UnixNano(), req.RequestName, name)

		web_concurrent_request(&web_request{
			RequestId:   requestId,
			RequestName: name,
			ThreadId:    thread_id,
			URL:         resourceUrl,
			Method:      "GET",
			Resource:    1,
			Referer:     extraRes.Referer,
		})
	}

	go_web_concurrent_end(0)

	return LR_PASS
}

func raw_header_string(res *http.Response) string {
	dump, err := httputil.DumpResponse(res, false)
	if err != nil {
		vlog("[warn] Unable to dump response headers", err)
		return ""
	}

	return string(dump)
}

func merge_param_matches(m1 map[string]bool, m2 map[string]bool) map[string]bool {
	out := m1
	for paramName, value := range m2 {
		existingValue, ok := out[paramName]
		if !ok {
			out[paramName] = value
		} else {
			out[paramName] = existingValue || value
		}
	}
	return out
}

func has_param_misses(m map[string]bool) bool {
	for _, b := range m {
		if !b {
			return true
		}
	}
	return false
}
